/* --------------------------------------------------------------------------
   WPPHP

   AUTHORS:
   Lukasz Karwacki <lukasm@wp-sa.pl>
   Wojciech Meller <wmeler@wp-sa.pl>

   -------------------------------------------------------------------------- */

#include "wpphp.h"

void wpphp_init(instance i, xmlnode x)
{
    wpphp wpp;
    xmlnode config;
	int k;

    log_debug(ZONE,"WPPHP initializing for section '%s'",i->id);

    /* create new transport instance */
    wpp = pmalloco(i->p,sizeof(_wpphp));
    wpp->i = i;
    wpp->xc = xdb_cache(i);

    config = xdb_get(wpp->xc,
					 jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:wpphp");
    
    if (config == NULL)
    {
        log_error(i->id,"WP PHP Configuration not found!");
        return;
    }

	wpp->script = xmlnode_get_tag_data(config,"script");
	if (!wpp->script)
	  wpp->script = pstrdup(i->p,"process.php");

	log_alert("config","using script %s",wpp->script);

	wpp->check_script = xmlnode_get_tag_data(config,"check_script");
	if (!wpp->check_script)
	  wpp->check_script = pstrdup(i->p,"process.php");

	log_alert("config","using check script %s",wpp->check_script);

	wpp->php_scripts = wpxhash_new(999);

	wpp->config = config;
	
	/* INIT PHP */
	jabber_php_init();

	wpp->threads = j_atoi(xmlnode_get_tag_data(config,"threads"),0);

	if ((wpp->threads > 50) || (wpp->threads < 1))
	  wpp->threads = 1;
			 
	log_alert("config","using %d PHP thread(s)",wpp->threads);

	for (k=1;k<wpp->threads;k++)
	  jabber_php_add_thread();

	//hash beat
	wpp->check_time = j_atoi(xmlnode_get_tag_data(config,"check"),0);
	if (wpp->check_time) {
	  log_alert("config","will run check script every %d sec",
				wpp->check_time);
	  
	  register_beat(wpp->check_time,
					wpphp_check,
					(void *)wpp);
	}

    register_phandler(i, o_DELIVER, wpphp_receive ,(void *) wpp);
    register_shutdown(wpphp_shutdown,(void *) wpp);
}

/* all Jabber packets are received here */
result wpphp_receive(instance i, dpacket d, void *arg) {
  wpphp wpp = (wpphp) arg;
  jpacket jp;
  php_packet pp;
  int packet=0;

  jp = NULL;

  switch(d->type) {
  case p_ROUTE: {
	return r_PASS;
  }			
  case p_NONE:
  case p_NORM:
	jp = jpacket_new(d->x);
	if ((!jp) || (jp->from == NULL)) {
	  xmlnode_free(jp->x); 
	  return r_DONE;
	}
	break;
  case p_XDB:
	packet = 1;
	break;
  case p_LOG:
	packet = 2;
	break;

  default:
	return r_ERR;
  }
  
  if (wpp->shutdown) {
	xmlnode_free(jp->x); 
	return r_DONE;
  }
		
  /* handle it in PHP , string must be copied */
  pp = pmalloco(xmlnode_pool(d->x),sizeof(_php_packet));
  pp->p = xmlnode_pool(d->x);
  pp->packet = xmlnode2str(d->x);
  pp->len = strlen(pp->packet);
  
  if (pp->len > MAX_JABBER_PACKET_LEN) {
	xmlnode_free(d->x); 
	return r_DONE;
  }

  pp->script = pstrdup(pp->p,wpp->script);
	
  if (packet == 0) {
	pp->to = jid_full(jp->to);
	pp->from = jid_full(jp->from);
	
	switch (jp->type) {
	case JPACKET_MESSAGE:
	  pp->type = PHP_MESSAGE_PACKET;
	  if (j_strcmp(xmlnode_get_attrib(jp->x,"type"),"chat")==0)
		pp->subtype = 1;
	  else
		pp->subtype = 0;
	  
	  pp->arg1 = xmlnode_get_tag_data(jp->x,"body");
	  pp->arg2 = xmlnode2str(xmlnode_get_tag(jp->x,"xhtml"));
	  break;
	case JPACKET_PRESENCE:
	  pp->type = PHP_PRESENCE_PACKET;

	  switch (jpacket_subtype(jp)) {
	  case JPACKET__AVAILABLE:
		pp->subtype = 1;
		break;
	  case JPACKET__UNAVAILABLE:
		pp->subtype = 2;
		break;
	  case JPACKET__PROBE:
		pp->subtype = 3;
		break;
	  case JPACKET__INVISIBLE:
		pp->subtype = 4;
		break;

		/* error */
	  default:
		pp->subtype = 0;
	  }

	  pp->arg1 = xmlnode_get_tag_data(jp->x,"show");
	  pp->arg2 = xmlnode_get_tag_data(jp->x,"status");


	  break;
	case JPACKET_IQ:
	  switch (jpacket_subtype(jp)) {
	  case JPACKET__GET:
		pp->subtype = 1;
		break;
	  case JPACKET__SET:
		pp->subtype = 2;
		break;
	  case JPACKET__RESULT:
		pp->subtype = 3;
		break;
	  }
	  pp->type = PHP_IQ_PACKET;
	  pp->arg1 = xmlnode_get_name(jp->iq);
	  pp->arg2 = jp->iqns;
	  break;
	case JPACKET_S10N:
	  pp->type = PHP_S10N_PACKET;

	  switch (jpacket_subtype(jp)) {
	  case JPACKET__SUBSCRIBE:
		pp->subtype = 1;
		break;
	  case JPACKET__SUBSCRIBED:
		pp->subtype = 2;
		break;
	  case JPACKET__UNSUBSCRIBE:
		pp->subtype = 3;
		break;
	  case JPACKET__UNSUBSCRIBED:
		pp->subtype = 4;
		break;
	  }
	  
	  break;
	default:
	  pp->type = 0;
	  break;
	}
  }
  /* xdb packet */
  else if (packet == 1) {
	pp->to = xmlnode_get_attrib(d->x, "to");
	pp->from = xmlnode_get_attrib(d->x, "from");
	pp->type = PHP_XDB_PACKET; /* xdb packet */
  }
  else  {
	pp->to = xmlnode_get_attrib(d->x, "to");
	pp->from = xmlnode_get_attrib(d->x, "from");
	pp->type = PHP_LOG_PACKET; /* log packet */
  }
  
  logger("packet_in",pp->from,pp->packet);
  
  jabber_php_queue_packet(pp);
  
  return r_DONE;
}

/* shutdown */
void wpphp_shutdown(void *arg)
{
  wpphp wpp = (wpphp) arg;
  
  log_alert(ZONE,"WP PHP shutting down");
  wpp->shutdown = 1;
  
  /* free PHP */
  jabber_php_shutdown();  
}

/* check */
result wpphp_check(void *arg)
{
  wpphp wpp = (wpphp) arg;
  int ret = 0;
  php_packet pp;
  pool p;

  if (wpp->shutdown)
	return r_DONE;

  /* check */
  if (wpp->check_changed) {
	wpp->check_changed = 0;

	if (wpp->check_time) {
	  register_beat(wpp->check_time, 
						 wpphp_check,
						 (void *)wpp);
	}
	ret = 1;
  }

  p = pool_heap(512);
  pp = pmalloco(p,sizeof(_php_packet));
  pp->p = p;
  pp->packet = pstrdup(p,"check");
  pp->len = strlen(pp->packet);
  
  pp->script = pstrdup(p,wpp->check_script);
  
  jabber_php_queue_packet(pp);  

  if (ret)
	return r_UNREG;
  else
	return r_DONE;
}
